# coding: utf-8
"""
Base classes to handle reusable code for S3 connections etc.
"""

from boto.s3.connection import S3Connection


class BaseS3ConnectionHandler(object):
    """
    Base class to derive from. Handles a connection to S3.
    Allows passing an existing connection for reuse.
    """

    def __init__(self, 
        connection, 
        aws_access_key_id=None, 
        aws_key_secret=None):
        self.connection = connection
        self.aws_access_key_id = aws_access_key_id
        self.aws_key_secret = aws_key_secret
        self._bucket = None # caches last used S3Bucket object


    def connect(self):
        """
        Creates a connection to S3 using boto.
        """
        if self.connection is not None:
            return
        self.connection = S3Connection(
            self.aws_access_key_id, 
            self.aws_key_secret)


    def get_bucket(self, bucket_name):
        """
        Returns a boto S3Bucket object.
        Caches it on the class for further reuse.
        """
        if not self._bucket or self._bucket.name != bucket_name:
            self._bucket = self.connection.get_bucket(bucket_name)
        return self._bucket



# coding: utf-8
"""
Default log file class representing a single log file as created by S3.
"""

import re
import datetime
from dateutil.parser import parse as dateutil_parse


LINE_REGEX_STRING  = r'(\S+) (\S+) \[(.*?)\] (\S+) (\S+) ' \
           r'(\S+) (\S+) (\S+) "([^"]+)" ' \
           r'(\S+) (\S+) (\S+) (\S+) (\S+) (\S+) ' \
           r'"([^"]+)" "([^"]+)" (\S+)'
LINE_REGEX_COMPILED = re.compile(LINE_REGEX_STRING)


class LogFile(object):
    """
    Base class to represent S3 log files.
    """    

    def __init__(self, key, prefix):
        self.key = key
        self._raw_contents = None
        self._entries = None
        object_name = key.name.replace(prefix, '')
        parts = object_name.split('-')
        self.log_id = parts.pop(-1)
        parts = map(lambda p: int(p), parts)
        self.creation_datetime = datetime.datetime(*parts)


    @property
    def raw_contents(self):
        """
        Property returning the exact S3 log file string contents.
        """
        if not self._raw_contents:
            self._raw_contents = self.key.get_contents_as_string()
        return self._raw_contents


    @property
    def entries(self):
        """
        List of LogFileEntry objects parsed from self.raw_contents.
        """
        if not self._entries:
            self._entries = [
                LogFileEntry(l, self) 
                for l in self.raw_contents.splitlines() ]
        return self._entries


    def delete(self):
        """
        Deletes the log file on S3 as long as we have a key object.
        """
        if self.key:
            self.key.delete()
            return True
        return False


class LogFileEntry(object):
    """
    A single entry (line) in an S3 log file.
    """

    FIELD_NAMES = ('bucket_owner', 'bucket', 'datetime', 'ip', 'requestor_id', 
    'request_id', 'operation', 'key', 'http_method_uri_proto', 'http_status', 
    's3_error', 'bytes_sent', 'object_size', 'total_time', 'turn_around_time',
    'referer', 'user_agent', 'version_id')

    def __init__(self, line, log_file):
        self.line = line
        # keep track of which log file this entry came from
        self.log_id = log_file.log_id
        self.parse()

    def parse(self):
        match = LINE_REGEX_COMPILED.match(self.line)
        values = [match.group(1+n) for n in range(18)]
        for (name, value) in zip(self.FIELD_NAMES, values):
            setattr(self, name, value)
        # fix : separating date and time instade of space
        # so that the default dateutil parser can parse it
        self.datetime = self.datetime.replace(':', ' ', 1)
        self.datetime = dateutil_parse(self.datetime)
        return self



# coding: utf-8
"""
Code used to delete S3 files.
"""

from .base import BaseS3ConnectionHandler
from .log_file import LogFile


class Deleter(BaseS3ConnectionHandler):
    """
    Simple class used to delete log files.
    """

    def __init__(self, *args, **kwargs):
        super(Deleter, self).__init__(*args, **kwargs)


    def delete_files(self, bucket_name, key_names, prefix=''):
        """
        Deletes a list of files from a bucket.
        """
        self.connect()
        bucket = self.get_bucket(bucket_name)
        for key_name in key_names:
            name = key_name \
                if key_name.startswith(prefix) \
                else prefix + key_name
            bucket.delete_key(name)



# coding: utf-8
"""
Tests for ls-s3-logs
"""

import unittest
import datetime

from testconfig import config

from lss3logs.download import Downloader
from lss3logs.delete import Deleter
from lss3logs.log_file import LogFile, LogFileEntry


class S3LogsTestCase(unittest.TestCase):
    """
    Tests everything there is about downloading log files
    """

    def __init__(self, *args, **kwargs):
        super(S3LogsTestCase, self).__init__(*args, **kwargs)
        self._downloader = None
        self._test_logs = None


    def setUp(self):
        """
        Setting up DownloaderTestCase
        """
        print 'setting up DownloaderTestCase test function'


    def _download_test_logs(self, force=False):
        """
        Downloads some log files to test with.
        This function is meant to be called by test functions.
        """
        if self._test_logs and not force:
            return self._test_logs
        aws = config['aws']
        self._downloader = Downloader(
            connection=None,
            aws_access_key_id=aws['access_key_id'],
            aws_key_secret=aws['key_secret'],
        )
        self._test_logs = self._downloader.download_files(
            aws['s3_bucket'],
            prefix=aws['logs_path'],
            max_logs=aws['max_logs_to_download'])
        return self._test_logs


    def test_downloading_logs(self):
        """
        Test downloading log files, 
        and getting back LogFile objects.
        """
        self._download_test_logs()
        self.assertTrue(
            len(self._test_logs) > 0, 
            'cannot test without logs')
        first_log = self._test_logs[0]
        self.assertIsInstance(
            first_log, LogFile, 
            'should return LogFile objects')
        self.assertIsInstance(
            first_log.log_id, unicode, 
            'log id should be a string')
        self.assertTrue(
            len(first_log.log_id) > 0, 
            'log id should not be an empty string')
        self.assertIsInstance(
            first_log.creation_datetime, datetime.datetime, 
            'log creation datetime should be a datetime object')


    def test_parsing_logs(self):
        """
        Test parsing log files, 
        and getting back LogFileEntry objects.
        """
        self._download_test_logs()
        entries = self._test_logs[0].entries
        self.assertTrue(
            len(entries) > 0, 
            'log should contain entries')
        self.assertIsInstance(
            entries[0], LogFileEntry, 
            'log entry should be a LogFileEntry object')


    def test_deleting_logs_using_keys(self):
        """
        Test deleting logs, checking that if 
        we download again we get different ones.
        This test uses keys still in memory from 
        when the logs were saved.
        """
        self._download_test_logs()
        self.assertTrue(
            len(self._test_logs) > 1, 
            'more than 1 log file are needed to test deletion')
        first_log_file = self._test_logs[0]
        first_log_id = first_log_file.log_id
        first_log_file.delete()
        self._download_test_logs(force=True)
        self.assertNotEqual(
            first_log_id,
            self._test_logs[0].log_id, 
            'first log file still present when it should have been deleted')


    def test_deleting_logs_using_deleter(self):
        """
        Test deleting logs, checking that if 
        we download again we get different ones.
        This test uses the Deleter class, for when 
        we don't have boto Key objects anymore.
        """
        self._download_test_logs()
        self.assertTrue(
            len(self._test_logs) > 1, 
            'more than 1 log file are needed to test deletion')
        first_log_file = self._test_logs[0]
        first_log_id = first_log_file.log_id

        # we use the key object's name directly here for convenience
        # but we can imagine it's been saved and retrieved 
        # from a database or such
        key_names = (first_log_file.key.name,)
        deleter = Deleter(self._downloader.connection)
        # import pdb; pdb.set_trace()
        deleter.delete_files(
            config['aws']['s3_bucket'],
            key_names,
            prefix=config['aws']['logs_path'])

        self._download_test_logs(force=True)
        self.assertNotEqual(
            first_log_id,
            self._test_logs[0].log_id, 
            'first log file still present when it should have been deleted')




import os
from setuptools import setup, find_packages


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ''


# Hack to prevent stupid "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when running `python
# setup.py test` (see
# http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass

required_packages = [
    'boto==2.8.0',
    'python-dateutil'
]

test_packages = [
    'pylint',
    'nose',
    'nose-testconfig==0.8',
    'coverage',
]

setup(
    name='ls-s3-logs',
    version=__import__('lss3logs').__version__,
    author='Liquid State Pty Ltd',
    author_email='dev@liquid-state.com',
    packages=find_packages(),
    url='https://github.com/liquid-state/ls-s3-logs',
    license='BSD',
    description=u' '.join(__import__('lss3logs').__doc__.splitlines()).strip(),
    classifiers=[
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
    ],
    long_description=read_file('README.rst'),
    install_requires=required_packages,
    tests_require=required_packages + test_packages,
    test_suite='nose.collector',
)


# coding: utf-8
"""
Classes used to download a bunch of logs at once and return LogFile objects.
"""

from .base import BaseS3ConnectionHandler
from .log_file import LogFile


class Downloader(BaseS3ConnectionHandler):
    """
    Default class used to download log files. Return sequences of LogFile objects.
    """

    def download_files(self, bucket_name, prefix='', max_logs=0):
        """
        Downloads all log files in a bucket at the given prefix
        """
        logs = [ ]
        self.connect()
        bucket = self.get_bucket(bucket_name)
        kwargs = { 'prefix': prefix }
        if max_logs > 0:
            kwargs['max_keys'] = max_logs
        keys = bucket.get_all_keys(**kwargs)
        for k in keys:
            logs.append(LogFile(k, prefix))
        return logs
